import hepi
import smpl
import hepi.resummino as rs
import hepi.input as input
#import hepi
import numpy as np
import matplotlib.pyplot as plt
import subprocess
import time
import os
import sys
from smpl import plot

user = "a_neuw01"
plt.rcParams.update({'font.size': 15})
plt.rc('legend', fontsize=12)


def get_job_name() -> str:
    return os.path.basename(sys.argv[0])

def wait():
    output = "\n\n\n"
    while(len(output.split('\n')) > 2):
        bashCommand = "squeue -u " + user + " --name=" + get_job_name() 
        if output != "\n\n\n":
            #print(bashCommand , " -> ", output)
            time.sleep(60)
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        output=output.decode()


rs.set_path("/home/a/"+user+"/git/resummino/")
input.set_output_dir("/scratch/tmp/"+user+"/hepi/")
input.set_input_dir("/home/a/"+user+"/input/")
input.set_pre("while [[ $(squeue -u " + user+ " | wc -l) -gt 1990 ]]; do sleep 60; done && sbatch --job-name=" +get_job_name() +
" --ntasks-per-node 1 --cpus-per-task 1  --time=3-00:00:00 --mem=100M --partition normal" +
" --mail-type=FAIL --mail-user="+user+"@uni-muenster.de")
print(rs.get_path())






from typing import List
import subprocess
from string import Template
import numpy as np
import pkgutil
from .. import Input, Result, LD2DL, get_output_dir, get_input_dir
import re
from uncertainties import ufloat_fromstr
import os.path
from pathlib import Path
from .result import MadgraphResult, is_valid, parse_single
import enlighten
import time
import difflib
import pyslha
from smpl.parallel import *
import hashlib

madgraph_path = "/opt/MG5_aMC_v2_7_0/"
"""madgraph folder"""


def set_path(p):
    """
    Set the path to the MadGraph folder containing the binary in './bin'.
    """
    global madgraph_path
    madgraph_path = p+ ("/" if p[-1]!="/" else "")


def get_path():
    """
    Returns the currently set MadGraph path.
    """
    global madgraph_path
    return madgraph_path


class RunParams:
    def __init__(self, dic,  skip=False,madstr=True):
        self.dic = dic
        self.skip = skip
        self.madstr = madstr


def run(params: List[Input], noskip=False,madstr=True,para=True):
    rps = _queue(params, noskip,madstr)
    _run(rps,para)
    outs = LD2DL(rps)["dic"]
    results = _parse(outs)
    rdl = LD2DL(results)
    pdl = LD2DL(params)
    return {**rdl, **pdl}


def _parse(outputs: List[str]) -> List[MadgraphResult]:
    rsl = []
    for r in par(lambda f: parse_single(f["out"]), outputs):
        rsl.append(r)
    return rsl
def namehash(n):
    m = hashlib.sha256()
    m.update(str(n).encode('utf-8'))
    return m.hexdigest()

def _queue(params: List[Input], noskip=False,madstr=True,para=True) -> List[RunParams]:
    Path("output").mkdir(parents=True, exist_ok=True)
    Path("input").mkdir(parents=True, exist_ok=True)
    ret = []
    for p in params:
        d = p.__dict__
        d["code"] = "MG"
        # TODO insert defautl if missing in d!
        name = namehash("_".join("".join(str(_[0]) + "_" + str(_[1]))
                        for _ in d.items()).replace("/", "-"))
        skip = False
        if not noskip and os.path.isfile(get_output_dir() + name + ".out") and is_valid(get_output_dir() + name + ".out",p,d):
            print("skip", end='')
            skip = True

        d["dir"] = get_output_dir() + name + ".dir"
        d["bdir"] = get_output_dir() + name + ".bdir"

        data = pkgutil.get_data(__name__, ["lo.mg", "nlo.mg"][p.order]).decode( 'utf-8')
        src = Template(data)
        result = src.substitute(d)
        open(get_input_dir() + name + ".mg", "w").write(result)
        if not skip:
            open(get_output_dir() + name + ".out", "w").write(result + "\n\n")

        mgfile = ["run_card_no_madstr.dat", "run_card_with_madstr.dat"][p.order]
        if not madstr:
            mgfile = "run_card_no_madstr.dat"
        data = pkgutil.get_data(__name__, mgfile).decode( 'utf-8')

        src = Template(data)
        result = src.substitute(d)
        open(get_input_dir() + name + ".dat", "w").write(result)
        if not skip:
            open(get_output_dir() + name + ".out", "a").write(result + "\n\n")

        sname = d['slha']
        with open(get_input_dir() + sname, 'r') as f:
            #src = Template(f.read())
            #result = src.substitute(d)
            #open(get_input_dir() + sname + ".in", "w").write(result)
            if not skip:
                open(get_output_dir() + name + ".out",
                     "a").write(f.read() + "\n\n")

        ret.append(RunParams({'in': get_input_dir()+name + ".mg",
                              'dir': d["dir"],
                              'bdir': d["bdir"],
                              'run': get_input_dir() + name + ".dat",
                              'slha': get_input_dir() + sname,
                              'out': get_output_dir()+name + ".out"}, skip,madstr))

    return ret


def _run(rps: List[RunParams],para=True):
    # TODO clean up on exit emergency
    global madgraph_path
    # TODO RS build path checks?!?!
    template =  \
        'rm -rf {dir} && cp -r ' + rps[0].dic["bdir"] + \
        ' {dir}  && cp {slha} {dir}/Cards/param_card.dat && cp {run} {dir}/Cards/run_card.dat && echo "nb_core = 1" >> {dir}/Cards/amcatnlo_configuration.txt && nice -n 5 {dir}/bin/calculate_xsect -f >> {out}'
    print(rps[0].dic["out"])
    if not rps[0].skip:
        mgcom = 'bin/mg5_aMC'
        if rps[0].madstr:
            mgcom = 'bin/mg5_aMC --mode="MadSTR"'
        com = madgraph_path + mgcom + \
            ' --file {in} >> {out} && cp {slha} {bdir}/Cards/param_card.dat && cp {run} {bdir}/Cards/run_card.dat && sed -i \'s/.*= req_acc_FO/ 1 = req_acc_FO/g\' {bdir}/Cards/run_card.dat && echo "automatic_html_opening = False" >> {bdir}/Cards/amcatnlo_configuration.txt && nice -n 5 {bdir}/bin/calculate_xsect -f'
        pp = subprocess.Popen(com.format(**rps[0].dic), shell=True)
        pp.wait()
    # Run commands in parallel
    processes = []

    for rp in rps:
        if not rp.skip:
            command = template.format(**rp.dic)
            # print(command)
            process = subprocess.Popen(command, shell=True)
            processes.append(process)
            if not para:
                process.wait()

    # Collect statuses
    output = [p.wait() for p in processes]
    return output


# print(sp)
# names = queue(sp)
# run(names)



from .util import LD2DL
import numpy as np
from typing import List
from uncertainties import unumpy
import uncertainties as unc
from smpl import plot
import lhapdf
import warnings

#numerical convergence should be better by a factor of 10 to avoid spoiling the scale/pdf uncertainties
required_numerical_uncertainty_factor = 10 

class Result:
    def __init__(self, lo, nlo, nlo_plus_nll):
        self.LO = lo
        self.NLO = nlo
        self.NLO_PLUS_NLL = nlo_plus_nll
        if lo is not None and lo != 0:
            self.K_LO = lo/lo
        else:
            self.K_LO = None
        #else:
        #    print("lo None or lo=",lo)
        if nlo is not None and lo != 0:
            self.K_NLO = nlo/lo
        else:
            self.K_NLO = None
        #else:
        #    print("nlo None or lo=",lo)
        if nlo_plus_nll is not None and lo != 0:
            self.K_NLO_PLUS_NLL = nlo_plus_nll/lo
        else:
            self.K_NLO_PLUS_NLL = None

        if nlo_plus_nll is not None and nlo != 0:
            self.NLO_PLUS_NLL_OVER_NLO = nlo_plus_nll/nlo
        else:
            self.NLO_PLUS_NLL_OVER_NLO  =None
        #else:
        #    print("nlo+nll None or lo=",lo)





def pdf_error(li, dl):
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [attr for attr in dir(example) if not callable(
        getattr(example, attr)) and not attr.startswith("__")]
    dl["LO_PDF"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_PDF_CENTRAL"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_PDF_ERRSYM"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PDF"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PDF_CENTRAL"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PDF_ERRSYM"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_CENTRAL"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_PDF_ERRSYM"] = np.array([None]*len(dl["pdfset_nlo"]))
    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0:
            set = lhapdf.getPDFSet(dl["pdf_nlo"][i])
            pdfs = [0.0] * set.size
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (dl[s][i] == dl[s][j]) and s != "pdfset_nlo" and s != "precision" and s!="max_iters":
                        same = False
                if same:
                    pdfs[dl["pdfset_nlo"][j]] = j

            dl["LO_PDF_CENTRAL"][i] = plot.unv(dl["LO"][i])
            dl["LO_PDF_ERRPLUS"][i] = 0.0
            dl["LO_PDF_ERRMINUS"][i] = 0.0
            dl["LO_PDF_ERRSYM"][i] = 0.0
            # lo_unc = set.uncertainty(
            #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            nlo_unc = set.uncertainty(
                [plot.unv(dl["NLO"][k]) for k in pdfs], 90)
            dl["NLO_PDF_CENTRAL"][i] = nlo_unc.central
            dl["NLO_PDF_ERRPLUS"][i] = nlo_unc.errplus
            dl["NLO_PDF_ERRMINUS"][i] = -nlo_unc.errminus
            dl["NLO_PDF_ERRSYM"][i] = nlo_unc.errsymm
            #TODO error sym to minus and plus
            if(plot.usd(dl["NLO"][i])*required_numerical_uncertainty_factor > dl["NLO_PDF_ERRPLUS"][i] or  plot.usd(dl["NLO"][i])*required_numerical_uncertainty_factor > -dl["NLO_PDF_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo precision vs pdf", RuntimeWarning)

            nlo_plus_nll_unc = set.uncertainty(
                [plot.unv(dl["NLO_PLUS_NLL"][k]) for k in pdfs], 90)
            dl["NLO_PLUS_NLL_PDF_CENTRAL"][i] = nlo_plus_nll_unc.central
            dl["NLO_PLUS_NLL_PDF_ERRPLUS"][i] = nlo_plus_nll_unc.errplus
            dl["NLO_PLUS_NLL_PDF_ERRMINUS"][i] = -nlo_plus_nll_unc.errminus
            dl["NLO_PLUS_NLL_PDF_ERRSYM"][i] = nlo_plus_nll_unc.errsymm
            #TODO error sym to minus and plus
            if(plot.usd(dl["NLO_PLUS_NLL"][i])*required_numerical_uncertainty_factor > dl["NLO_PLUS_NLL_PDF_ERRPLUS"][i] or  plot.usd(dl["NLO_PLUS_NLL"][i])*required_numerical_uncertainty_factor > -dl["NLO_PLUS_NLL_PDF_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo_plus_nll precision vs pdf", RuntimeWarning)


    mask = dl["LO_PDF_CENTRAL"]!= np.array(None)
    dl["LO_PDF"][mask] = unumpy.uarray(plot.unv(dl["LO_PDF_CENTRAL"][mask])+dl["LO_PDF_ERRPLUS"][mask]/2.+dl["LO_PDF_ERRMINUS"][mask]/2.,
        (+dl["LO_PDF_ERRPLUS"][mask]-dl["LO_PDF_ERRMINUS"][mask])/2.)

    mask = dl["NLO_PDF_CENTRAL"]!= np.array(None)
    dl["NLO_PDF"][mask] = unumpy.uarray(plot.unv(dl["NLO_PDF_CENTRAL"][mask])+dl["NLO_PDF_ERRPLUS"][mask]/2.+dl["NLO_PDF_ERRMINUS"][mask]/2.,
        (+dl["NLO_PDF_ERRPLUS"][mask]-dl["NLO_PDF_ERRMINUS"][mask])/2.)

    mask = dl["NLO_PLUS_NLL_PDF_CENTRAL"]!= np.array(None)
    dl["NLO_PLUS_NLL_PDF"][mask] = unumpy.uarray(plot.unv(dl["NLO_PLUS_NLL_PDF_CENTRAL"][mask])+dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask]/2.+dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask]/2.,
        (+dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask]-dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask])/2.)
    return dl


def scale_error(li, dl):
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [attr for attr in dir(example) if not callable(
        getattr(example, attr)) and not attr.startswith("__")]
    dl["LO_SCALE"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_SCALE_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_SCALE_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_SCALE"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_SCALE_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_SCALE_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_SCALE_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    for i in range(len(dl["pdfset_nlo"])):
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0:
            scales = []
            for j in range(len(dl["pdfset_nlo"])):
                same = True
                for s in members:
                    if not (dl[s][i] == dl[s][j]) and s != "mu_f" and s != "mu_r" and s != "precision" and s!="max_iters":
                        same = False
                if same:
                    scales.append(j)

            # lo_unc = set.uncertainty(
            #    [plot.unv(dl["LO"][k]) for k in pdfs], -1)
            dl["LO_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["LO"][k]) for k in scales])-plot.unv(dl["LO"][i])
            dl["LO_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["LO"][k]) for k in scales])-plot.unv(dl["LO"][i])
            if(plot.usd(dl["LO"][i])*required_numerical_uncertainty_factor > dl["LO_SCALE_ERRPLUS"][i] or  plot.usd(dl["LO"][i])*required_numerical_uncertainty_factor > -dl["LO_SCALE_ERRMINUS"][i]):
                warnings.warn("too low numerical lo precision vs scale", RuntimeWarning)

            dl["NLO_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["NLO"][k]) for k in scales])-plot.unv(dl["NLO"][i])
            dl["NLO_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["NLO"][k]) for k in scales])-plot.unv(dl["NLO"][i])
            if(plot.usd(dl["NLO"][i])*required_numerical_uncertainty_factor > dl["NLO_SCALE_ERRPLUS"][i] or  plot.usd(dl["NLO"][i])*required_numerical_uncertainty_factor > -dl["NLO_SCALE_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo precision vs scale", RuntimeWarning)

            dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][i] = np.max(
                [plot.unv(dl["NLO_PLUS_NLL"][k]) for k in scales])-plot.unv(dl["NLO_PLUS_NLL"][i])
            dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][i] = np.min(
                [plot.unv(dl["NLO_PLUS_NLL"][k]) for k in scales])-plot.unv(dl["NLO_PLUS_NLL"][i])
            if(plot.usd(dl["NLO_PLUS_NLL"][i])*required_numerical_uncertainty_factor > dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][i] or  plot.usd(dl["NLO_PLUS_NLL"][i])*required_numerical_uncertainty_factor > -dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][i]):
                warnings.warn("too low numerical nlo_plus_nll precision vs scale", RuntimeWarning)

    mask = dl["LO_SCALE_ERRPLUS"]!= np.array(None)
    dl["LO_SCALE"][mask] = unumpy.uarray(plot.unv(dl["LO"][mask])+dl["LO_SCALE_ERRPLUS"][mask]/2.+dl["LO_SCALE_ERRMINUS"][mask]/2.,
        (+dl["LO_SCALE_ERRPLUS"][mask]-dl["LO_SCALE_ERRMINUS"][mask])/2.)

    mask = dl["NLO_SCALE_ERRPLUS"]!= np.array(None)
    dl["NLO_SCALE"][mask] = unumpy.uarray(plot.unv(dl["NLO"][mask])+dl["NLO_SCALE_ERRPLUS"][mask]/2.+dl["NLO_SCALE_ERRMINUS"][mask]/2.,
        (+dl["NLO_SCALE_ERRPLUS"][mask]-dl["NLO_SCALE_ERRMINUS"][mask])/2.)

    mask = dl["NLO_PLUS_NLL_SCALE_ERRPLUS"]!= np.array(None)
    dl["NLO_PLUS_NLL_SCALE"][mask] = unumpy.uarray(plot.unv(dl["NLO_PLUS_NLL"][mask])+dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask]/2.+dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask]/2.,
        (+dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask]-dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask])/2.)
    
    return dl

def combine_errors(dl):
    dl["LO_NOERR"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["LO_COMBINED"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_NOERR"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_COMBINED"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_NOERR"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_ERRPLUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_ERRMINUS"] = np.array([None]*len(dl["pdfset_nlo"]))
    dl["NLO_PLUS_NLL_COMBINED"] = np.array([None]*len(dl["pdfset_nlo"]))


    mask = dl["LO_PDF_CENTRAL"]!= np.array(None)
    dl["LO_NOERR"][mask]= plot.unv(dl["LO"][mask]).astype(float)
    dl["LO_ERRPLUS"][mask]= np.sqrt(dl["LO_PDF_ERRPLUS"][mask].astype(float)**2+dl["LO_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["LO_ERRMINUS"][mask]= -np.sqrt(dl["LO_PDF_ERRMINUS"][mask].astype(float)**2+dl["LO_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["LO_COMBINED"][mask] = unumpy.uarray(plot.unv(dl["LO"][mask])+dl["LO_ERRPLUS"][mask]/2.+dl["LO_ERRMINUS"][mask]/2.,
        +dl["LO_ERRPLUS"][mask]-dl["LO_ERRMINUS"][mask])

    mask = dl["NLO_PDF_CENTRAL"]!= np.array(None)
    dl["NLO_NOERR"][mask]= plot.unv(dl["NLO"][mask]).astype(float)
    dl["NLO_ERRPLUS"][mask]= np.sqrt(dl["NLO_PDF_ERRPLUS"][mask].astype(float)**2+dl["NLO_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["NLO_ERRMINUS"][mask]= -np.sqrt(dl["NLO_PDF_ERRMINUS"][mask].astype(float)**2+dl["NLO_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["NLO_COMBINED"][mask] = unumpy.uarray(plot.unv(dl["NLO"][mask])+dl["NLO_ERRPLUS"][mask]/2.+dl["NLO_ERRMINUS"][mask]/2.,
        +dl["NLO_ERRPLUS"][mask]-dl["NLO_ERRMINUS"][mask])

    mask = dl["NLO_PLUS_NLL_PDF_CENTRAL"]!= np.array(None)
    dl["NLO_PLUS_NLL_NOERR"][mask]= plot.unv(dl["NLO_PLUS_NLL"][mask]).astype(float)
    dl["NLO_PLUS_NLL_ERRPLUS"][mask]= np.sqrt(dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask].astype(float)**2+dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask].astype(float)**2)
    dl["NLO_PLUS_NLL_ERRMINUS"][mask]= -np.sqrt(dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask].astype(float)**2+dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask].astype(float)**2)
    dl["NLO_PLUS_NLL_COMBINED"][mask] = unumpy.uarray(plot.unv(dl["NLO_PLUS_NLL"][mask])+dl["NLO_PLUS_NLL_ERRPLUS"][mask]/2.+dl["NLO_PLUS_NLL_ERRMINUS"][mask]/2.,
        +dl["NLO_PLUS_NLL_ERRPLUS"][mask]-dl["NLO_PLUS_NLL_ERRMINUS"][mask])

    return dl



from distutils import command
from typing import List
import warnings
import pyslha
from hepi import get_input_dir, get_output_dir
#from hepi.input import slha_scan
import subprocess
import os
from hepi.input import Input, update_slha

spheno_path = "~/git/SPheno-3.3.8/"
"""spheno folder containing the binary in './bin'."""


def set_path(p):
    """
    Set the path to the SPheno folder containing the binary in './bin'.
    """
    global spheno_path
    spheno_path = p + ("/" if p[-1]!="/" else "")


def get_path():
    """
    Returns the currently set SPheno path.
    """
    global spheno_path
    return spheno_path


def run(slhas : List[Input]) -> List[Input]:
    """
    Run the passed list of parameters for SPheno.
    """
    if os.path.exists("Messages.out"):
        os.remove("Messages.out")
    for s in slhas:
        #print(s.slha)
        comm= "cp " +  get_input_dir() + s.slha  + " spheno_tmp.in && " +get_path() + "bin/SPheno spheno_tmp.in && mv "  + "SPheno.spc " + get_input_dir() + s.slha  + " && sed -i '/Created/d' " + get_input_dir() + s.slha
        proc = subprocess.Popen(comm, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc.wait()
        update_slha(s)
    if os.path.exists("Messages.out"):
        with open("Messages.out",'r') as r:
            t = r.read()
            if t != "":
                warnings.warn(r.read())
    return slhas



from hepi.input import Order
from .. import Input, Result, LD2DL, get_output_dir, get_input_dir
import re
from uncertainties import ufloat_fromstr


class MadgraphResult(Result):
    """
    MadGraph Result Data.
    """
    def __init__(self, lo, nlo):
        Result.__init__(self, lo, nlo, nlo)


def is_valid(file:str,p:Input,d):
    """
    Verifies that an file is a complete output.
    """
    order = p.order
    res = parse_single(file)
    if res.LO is not None and order is Order.LO:
        return True
    if res.LO is not None and res.NLO is not None and order is Order.NLO:
        return True
    if res.LO is not None and res.NLO is not None and res.NLO_PLUS_NLL is not None and order is Order.NLO_PLUS_NLL:
        return True
    print("RESTART" ,res.LO, res.NLO,res.NLO_PLUS_NLL, file)
    return False

def parse_single(file) -> MadgraphResult:
    # TODO generalize units like RS
    lo_pattern = re.compile(r'^\s*Total cross section:\s*(.*) pb')

    lo_result = None
    with open(file) as output:
        for line in output:
            tmp = lo_pattern.search(line)
            if tmp is not None:
                lo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))

    return MadgraphResult(lo_result, lo_result)



from scipy import interpolate

from sklearn.metrics import auc
import matplotlib as mpl
from smpl import plot as splot
import smpl
import numpy as np
import uncertainties as unc
from scipy.interpolate import make_interp_spline, BSpline

from uncertainties import unumpy
import matplotlib.pyplot as plt

from particle import PDGID
import pyslha
from particle import Particle
from particle.converters.bimap import DirectionalMaps

import matplotlib.cm as cm
from matplotlib import colors

from .input import Input, get_input_dir
from .util import get_name
from matplotlib.ticker import ScalarFormatter, NullFormatter
from smpl import io
from typing import List

def tex_table(dict_list,key,fname,scale=True,pdf=True,yscale=1.):
 dl = dict_list
 mask = dl["NLO_SCALE"]!= np.array(None)
 lo = splot.unv(dl["LO"][mask])
 nlo = splot.unv(dl["NLO"][mask])
 nlo_plus_nll = splot.unv(dl["NLO_PLUS_NLL"][mask])
 with open(fname,'w+') as f:
     for i in range(len(dl["LO"][mask])):
        f.write(
            "$" + io.gf(4).format(dl[key][mask][i]) + "$ \t&\t $"+
            "{:.3f}".format(lo[i]*yscale) 
                + "^{+"+"{:.1f}".format(dl["LO_SCALE_ERRPLUS"][mask][i]/lo[i]*100.)
                + "\%}_{" +"{:.1f}".format(dl["LO_SCALE_ERRMINUS"][mask][i]/lo[i]*100.)
                +  "\%}$ \t&\t $"+
            "{:.3f}".format(nlo[i]*yscale) 
                + "^{+"+"{:.1f}".format(dl["NLO_SCALE_ERRPLUS"][mask][i]/nlo[i]*100.)
                + ("\%+"+"{:.1f}".format(dl["NLO_PDF_ERRPLUS"][mask][i] /nlo[i]*100.) if pdf else "")
                + "\%}_{" +"{:.1f}".format(dl["NLO_SCALE_ERRMINUS"][mask][i]/nlo[i]*100.)
                + ("\%"+"{:.1f}".format(dl["NLO_PDF_ERRMINUS"][mask][i]/nlo[i]*100.) if pdf else "")
                +  "\%}$ \t&\t $"+
            "{:.3f}".format(nlo_plus_nll[i]*yscale) 
                + "^{+"+"{:.1f}".format(dl["NLO_PLUS_NLL_SCALE_ERRPLUS"][mask][i]/nlo_plus_nll[i]*100.)
                + ("\%+"+"{:.1f}".format(dl["NLO_PLUS_NLL_PDF_ERRPLUS"][mask][i]/nlo_plus_nll[i]*100.) if pdf else "")
                + "\%}_{" +"{:.1f}".format(dl["NLO_PLUS_NLL_SCALE_ERRMINUS"][mask][i]/nlo_plus_nll[i]*100.)
                + ("\%"+"{:.1f}".format(dl["NLO_PLUS_NLL_PDF_ERRMINUS"][mask][i]/nlo_plus_nll[i]*100.) if pdf else "")
                +  "\%}$ "+
            "\\\\\n"
        )

def title(axe,i:Input,scenario="",diff_L_R=None,extra="",**kwargs):
    axe.set_title(
        "$pp\\to"+get_name(i.particle1)+get_name(i.particle2) + "$"
        #+" at $\\sqrt{S} = " +str(i.energy/1000) + "$ TeV"
        +" for " +(i.slha.split(".")[0] if scenario =="" else scenario)
        #+" with " + i.pdf_nlo
        + " " + extra
    )



def energy_plot(dict_list, y, yscale=1.,xaxis="E [GeV]",yaxis="$\\sigma$ [pb]",label=None,**kwargs):
    plot(dict_list, "energy", y, label=label,
         xaxis=xaxis, yaxis=yaxis, logy=True, yscale=yscale,**kwargs)


def combined_energy_plot(dict_list,t,**kwargs):
    dl = dict_list
    mask = dl[t+"_pdf_central"]!= np.array(None)
    if 'axes' in kwargs and kwargs['axes'] is not None:
        color = next(kwargs['axes']._get_lines.prop_cycler)['color']
    else:
        color = next(plt.gca()._get_lines.prop_cycler)['color']
 
    splot.data(dict_list["energy"][mask],splot.unv(dict_list[t][mask]),
        xaxis="E [GeV]", yaxis="$\\sigma$ [pb]", fmt=".",logy=True, label=t,data_color=color)
    splot.data(dict_list["energy"][mask],dict_list[t+ "_scale"][mask],
        xaxis="E [GeV]", yaxis="$\\sigma$ [pb]", fmt=" ",logy=True,data_color=color)
    splot.data(dict_list["energy"][mask],dict_list[t+ "_combined"][mask],
        xaxis="E [GeV]", yaxis="$\\sigma$ [pb]", fmt=" ",logy=True,data_color=color,capsize=None)

def combined_plot(func,dict_list,t,*args,label=None,fill = False,fmt=".",interpolate=True,**kwargs):
    dl = dict_list
    mask = dl[t+"_pdf_central"]!= np.array(None)

    if 'axes' in kwargs and kwargs['axes'] is not None:
        color = next(kwargs['axes']._get_lines.prop_cycler)['color']
    else:
        color = next(plt.gca()._get_lines.prop_cycler)['color']
    func(dict_list,t+ "_noerr",*args,
         label=t if label is None else label,data_color=color,fill=False,interpolate=interpolate,mask = mask,**kwargs)
    func(dict_list,t+ "_scale",*args,
        fmt=" ",interpolate=False,data_color=color,mask = mask,label="",fill=False,**kwargs)
    func(dict_list,t+ "_combined",*args,
         fmt=" ",interpolate=False,data_color=color,capsize=None,label="",mask = mask,fill=fill,**kwargs)



def get_mass(l, id):
    ret = []
    for s in l["slha"]:
        d = pyslha.read(get_input_dir() + s)
        ret.append(d.blocks["MASS"][abs(id)])
    return np.array(ret)

def mass_plot(dict_list,  y, part, logy=True, yaxis="$\\sigma$ [pb]", yscale=1., label=None,**kwargs):
    dict_list["mass_"+ str(part)] = get_mass(dict_list, abs(part))
    plot(dict_list, "mass_" + str(part), y, label=label,
         xaxis="$m_{"+get_name(part) + "}$ [GeV]", yaxis=yaxis, logy=logy, yscale=yscale,**kwargs)


def mass_vplot(dict_list,  y,part, logy=True, yaxis="$\\sigma$ [pb]", yscale=1., label=None,mask=None,**kwargs):
  vplot(get_mass(dict_list,part)[mask], y[mask], label=label,
          xaxis="$m_{"+get_name(part) + "}$ [GeV]", yaxis=yaxis, logy=logy, yscale=yscale,mask=mask,**kwargs)
#    vplot(dict_list["mass_" + str(part)][mask], y[mask], label=label,
#          xaxis="$M_{"+get_name(part) + "}$ [GeV]", yaxis=yaxis, logy=logy, yscale=yscale,mask=mask,**kwargs)


def plot(dict_list, x, y, label=None, xaxis="E [GeV]", yaxis="$\\sigma$ [pb]",ratio=False,K=False,K_plus_1=False, logy=True, yscale=1.,mask=None,**kwargs):
    # TODO use kwargs
    if label is None:
        label = y
    if label == "":
        label = None

    vx = dict_list[x][mask]
    vy = dict_list[y][mask]

    if K:
        yaxis = "$K$"
        yscale = 1.0
        vy = vy / splot.unv(dict_list["LO"][mask])
        if K_plus_1:
            vy = vy + vy/vy
    if ratio:
        yaxis = "Ratio"
        yscale = 1.0
        vy = vy / splot.unv(vy)


    vplot(vx, vy, label, xaxis, yaxis, logy, yscale,mask=mask,**kwargs)

def index_open(var,idx):
    if len(idx) == 1:
        return var[idx]
    return index_open(var[idx[0]],idx[1:])

def slha_data(li,index_list):
    vx = []
    for l in li:
        b = pyslha.read(get_input_dir() + l.slha)
        vx.append(index_open(b.blocks,index_list))
    return np.array(vx)

def slha_plot(li,x,y,**kwargs):
    vx = slha_data(li,x)
    vy = slha_data(li,y)

    vplot(np.array(vx),np.array(vy),**kwargs)

def vplot(x, y, label=None, xaxis="E [GeV]", yaxis="$\\sigma$ [pb]", logy=True, yscale=1.,interpolate=True,plot_data=True,data_color=None,mask=-1,fill =False,data_fmt=".",fmt="-",**kwargs):
    color = data_color
    if label is None:
        #label = "??"
        pass
    if mask is None:
        x = x[0]
        y = y[0]
    vx = x
    vy = y

    xnew = np.linspace(vx[0], vx[-1], 300,)
    if interpolate:
        #print(vx,vy)
        spl = make_interp_spline(
            vx, splot.unv(vy), k=3)  # type: BSpline
        power_smooth = spl(xnew)
    if fill:
        spl_up = make_interp_spline(
                vx, splot.unv(vy)+splot.usd(vy), k=3)  # type: BSpline
        power_up_smooth = spl_up(xnew)
        spl_down = make_interp_spline(
                vx, splot.unv(vy)-splot.usd(vy), k=3)  # type: BSpline
        power_down_smooth = spl_down(xnew)
    if data_color is None:
        if 'axes' in kwargs and kwargs['axes'] is not None:
            bl, = kwargs['axes'].plot([], [])
        else:
            bl, = plt.gca().plot([], [])
        color = bl.get_color()
    if plot_data:
        splot.data(vx, vy*yscale, label=label, xaxis=xaxis, yaxis=yaxis,logy=logy, data_color=color,fmt=data_fmt, **kwargs)
    if interpolate:
        kargs = {}
        if not plot_data:
            kargs = {'xaxis':xaxis, 'yaxis':yaxis,'label':label}
        print('computed AUC using sklearn.metrics.auc: {}'.format(auc(xnew,power_smooth*yscale)))
        splot.data(xnew, power_smooth*yscale, logy=logy, fmt=fmt
              , init=False, data_color=color,  **kargs,**kwargs)
    if fill:
        plt.fill_between(xnew,power_up_smooth*yscale,power_down_smooth*yscale,alpha=0.3,color=color)
    if((np.any(np.less(vy,0)) or ( interpolate and np.any(np.less(power_smooth, 0)))) and logy):
        if plot_data:
            splot.data(vx, -vy*yscale,label="-"+label,xaxis=xaxis, yaxis=yaxis, logy=logy, data_color=color,fmt=data_fmt, **kwargs)
        if interpolate:
            if not plot_data:
                kargs = {'xaxis':xaxis, 'yaxis':yaxis,'label':"-"+label}
            splot.data(xnew, -power_smooth*yscale, logy=logy, fmt=None,linestyle=(0, (3, 1, 3, 1, 1, 1)),
                    init=False, data_color=color, **kargs,**kwargs)


def mass_mapplot(dict_list, part1, part2, z, logz=True, zaxis="$\\sigma$ [pb]", zscale=1., label=None):
    mapplot(dict_list, "mass_" + str(part1), "mass_" + str(part2), z,
            xaxis="$M_{"+get_name(part1) + "}$ [GeV]", yaxis="$M_{"+get_name(part2) + "}$ [GeV]", zaxis=zaxis, logz=logz, zscale=zscale)


def mapplot(dict_list, x, y, z, xaxis=None, yaxis=None, zaxis=None,**kwargs):
    if xaxis is None:
        xaxis = x
    if yaxis is None:
        yaxis = y
    if zaxis is None:
        zaxis = z
    vx = dict_list[x]
    vy = dict_list[y]
    vz = dict_list[z]
    map_vplot(vx,vy,vz,xaxis=xaxis,yaxis=yaxis,zaxis=zaxis,**kwargs)

def map_vplot(vx,vy,vz, xaxis=None, yaxis=None, zaxis=None, logz=True, zscale=1.):
    s = 1
    while vy[s] == vy[s-1]:
        s = s+1
    if s == 1:
        #print("flipped x y ")
        while vx[s] == vx[s-1]:
            s = s+1
        if s == 1:
            print("error too small map")
            return
        #x, y = y, x
        xaxis ,yaxis = yaxis,xaxis
        vx, vy = vy, vx

    grid = splot.unv(vz).reshape((int(np.rint(np.size(vx)/s)), s))*zscale
    if(logz):
        plt.imshow(grid, origin="lower", extent=(vx.min(), vx.max(),
                                                 vy.min(), vy.max()), cmap=cm.gist_heat, interpolation='nearest', norm=colors.LogNorm())
    else:
        plt.imshow(grid, origin="lower", extent=(vx.min(), vx.max(),
                                                 vy.min(), vy.max()), cmap=cm.gist_heat, interpolation='nearest')
    cb = plt.colorbar()
    cb.set_label(zaxis)
    plt.xlabel(xaxis)
    plt.ylabel(yaxis)
    plt.show()


fig = None
axs = None
lines = []
labels = []


def err_plt(axes,x,y,label=None,error=False):
    v= label
    ind = np.argsort(splot.unv(x))
    if error:
        l, _, _ = axes.errorbar(x[ind], splot.unv(y)[ind], yerr=splot.usd(y), capsize=5, label=v)
        return l
    else:
        l = axes.plot(x[ind], splot.unv(y)[ind], label=v)
        return l[0]

def scale_plot(dict_list, vl, seven_point_band=False, cont=False,error=True,li=None,plehn_color=False,yscale=1.,unit="pb",**kwargs):
    global fig, axs,lines,labels
    cycle_safe = mpl.rcParams['axes.prop_cycle'] 
    if plehn_color:
        mpl.rcParams['axes.prop_cycle'] = mpl.cycler(color=["b", "r", "g"]) 
    if not cont:
        fig, axs = plt.subplots(1, 5, figsize=(12, 3), sharey=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(wspace=0)
        if li is not None:
            title(axs[2],li[0],**kwargs)
   

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]





    if not cont:
        lines = []
        labels = []
        axs[0].plot([], [], ' ', color='k',label="$\mu_R=" + "\mu_F$")
        axs[1].plot([], [], ' ', color='k',label="$\mu_R=" + str(np.max(mr)) + "\mu_0$")
        axs[2].plot([], [], ' ', color='k',label="$\mu_F=" + str(np.min(mf)) + "\mu_0$")
        axs[3].plot([], [], ' ', color='k',label="$\mu_R=" + str(np.min(mr)) + "\mu_0$")
        axs[4].plot([], [], ' ', color='k',label="$\mu_F=" + str(np.max(mf)) + "\mu_0$")

    for v in vl:
        mv = dict_list[v]*yscale
        if seven_point_band:
            #mask = (mf/mr < 4.) & (mf/mr > 1./4.) & (mf <= 2) & (mf >= 1./2.) & (mr <= 2) & (mr >= 1./2.)
            mask =  (
                ((mf == 2.) & (mr == 2.)) | 
                 (   (mf == 2.) & (mr == 1.))  |            
                 (   (mf == 1.) & (mr == 1.))  |            
                 (   (mf == 1.) & (mr == 2.) ) |            
                (    (mf == 1/2.) & (mr == 1/2.)) | 
                (    (mf == 1/2.) & (mr == 1.)  )|            
                (    (mf == 1.) & (mr == 1/2.) ) 
            )
            mvmax = np.max(splot.unv(mv[mask]))
            mvmin = np.min(splot.unv(mv[mask]))

        mask = mf == mr
        l = err_plt(axs[0],mf[mask],mv[mask],error=error)
        #l, _, _ = axs[0].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5, label=v)
        if seven_point_band:
            axs[0].fill_between(mf[mask], mvmax, mvmin,
                                facecolor=l.get_color(), alpha=0.3)

        mask = mr == np.max(mr)
        l = err_plt(axs[1],mf[mask],mv[mask],error=error)
        lines.append(l)
        labels.append("$\\sigma_{\\mathrm{"+v.replace("NLO_PLUS_NLL","NLO+NLL").replace(" ","\\ ")+"} }$")
        #l, _, _ = axs[1].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[1].fill_between(mf[mask], mvmax, mvmin,
                                facecolor=l.get_color(), alpha=0.3)

        mask = mf == np.min(mf)
        l = err_plt(axs[2],mr[mask],mv[mask],error=error)
        #l, _, _ = axs[2].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[2].fill_between(mr[mask], mvmax, mvmin,
                                facecolor=l.get_color(), alpha=0.3)

        mask = mr == np.min(mr)
        l = err_plt(axs[3],mf[mask],mv[mask],error=error)
        #l, _, _ = axs[3].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[3].fill_between(mf[mask], mvmax, mvmin,
                                facecolor=l.get_color(), alpha=0.3)

        mask = mf == np.max(mf)
        l = err_plt(axs[4],mr[mask],mv[mask],error=error)
        #l, _, _ = axs[4].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            f = axs[4].fill_between(mr[mask], mvmax, mvmin,
                                facecolor=l.get_color(), alpha=0.3)
            lines.append(f)
            labels.append("$\\Delta \\sigma_{\\mathrm{" + v.replace("NLO_PLUS_NLL","NLO+NLL").replace(" ","\\<space>") + "} }$")


    axs[0].set_ylabel("$\sigma$ ["+unit+"]")

    axs[0].set_xscale("log")
    axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[0].set_xlabel("$\mu_{R,F}/\mu_0$")

    # axs[1].plot(t, s2)
    axs[1].set_xscale("log")
    axs[1].set_xlim(np.max(mf), np.min(mf))
    axs[1].set_xticks([1.])
    axs[1].xaxis.set_minor_formatter(NullFormatter())
    axs[1].set_xlabel("$\mu_{F}/\mu_0$")

    # axs[2].plot(t, s3)
    axs[2].set_xscale("log")
    axs[2].set_xlim(np.max(mf), np.min(mf))
    axs[2].set_xticks([1.])
    axs[2].xaxis.set_minor_formatter(NullFormatter())
    axs[2].set_xlabel("$\mu_{R}/\mu_0$")

    # axs[3].plot(t, s3)
    axs[3].set_xscale("log")
    axs[3].set_xlim(np.min(mf), np.max(mf))
    axs[3].set_xticks([1.])
    axs[3].xaxis.set_minor_formatter(NullFormatter())
    axs[3].set_xlabel("$\mu_{F}/\mu_0$")

    # axs[4].plot(t, s3)
    axs[4].set_xscale("log")
    axs[4].set_xlim(np.min(mf), np.max(mf))
    axs[4].set_xticks([1.])
    axs[4].xaxis.set_minor_formatter(NullFormatter())
    axs[4].set_xlabel("$\mu_{R}/\mu_0$")

    axs[0].legend(handletextpad=-0.0, handlelength=0,fancybox=False)
    axs[1].legend(handletextpad=-0.0, handlelength=0,fancybox=False)
    axs[2].legend(handletextpad=-0.0, handlelength=0,fancybox=False)
    axs[3].legend(handletextpad=-0.0, handlelength=0,fancybox=False)
    axs[4].legend(handletextpad=-0.0, handlelength=0,fancybox=False)
    fig.legend()
    fig.legends = []
    fig.legend(handles=lines,    labels=labels,   loc="center right")
    plt.subplots_adjust(right=0.84)
    # plt.show()
    mpl.rcParams['axes.prop_cycle'] = cycle_safe

    

def central_scale_plot(dict_list, vl, cont=False,error=True,yscale=1.,unit="pb"):
    global fig, axs
    if not cont:
        fig, axs = plt.subplots(3, 1, figsize=(12, 8), sharex=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(hspace=0)

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]

    for v in vl:
        mv = dict_list[v]*yscale

        mask = mf == mr
        l = err_plt(axs[0],mf[mask],mv[mask],label=v,error=error)
  

        mask = mf == 1.0
        l = err_plt(axs[1],mr[mask],mv[mask],error=error)


        mask = mr == 1.0
        l = err_plt(axs[2],mf[mask],mv[mask],error=error)



    if not cont:
        axs[0].plot([], [], ' ', label="$\mu_R=\mu_F=\mu$")
        axs[1].plot([], [], ' ', label="$\mu_F=\mu_0$, $\mu_R=\mu$")
        axs[2].plot([], [], ' ', label="$\mu_R=\mu_0$, $\mu_F=\mu$")

    axs[1].set_ylabel("$\sigma$ ["+unit+"]")

    axs[0].set_xscale("log")
    #axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[2].set_xlabel("$\mu/\mu_0$")

    # axs[1].plot(t, s2)
    #axs[1].set_xscale("log")
    #axs[1].set_xlim(np.max(mf), np.min(mf))
    #axs[1].set_xticks([1.])
    #axs[1].xaxis.set_minor_formatter(NullFormatter())
    #axs[1].set_xlabel("$\mu_{F}/\mu_0$")

    # axs[2].plot(t, s3)
    #axs[2].set_xscale("log")
    #axs[2].set_xlim(np.max(mf), np.min(mf))
    #axs[2].set_xticks([1.])
    #axs[2].xaxis.set_minor_formatter(NullFormatter())
    #axs[2].set_xlabel("$\mu_{R}/\mu_0$")

    axs[0].legend()
    axs[1].legend()
    axs[2].legend()
    # plt.show()


def mass_and_K_plot(dl,li,p,scale=False,pdf=False,plehn=True,combined=False,cont = False,figsize=(6,8),**kwargs):
    global fig, axs
    if not cont:
        fig, axs = plt.subplots(2, 1, figsize=figsize, sharex=True, gridspec_kw={'height_ratios': [3, 1]})
        # Remove horizontal space between axes
        fig.subplots_adjust(hspace=0)
        title(axs[0],li[0],**kwargs)
    if combined:
        for i in [0,1]:
            kargs = {'logy' : [True,False][i], 'interpolate' : False,'axes':axs[i],'K':[False,True][i],'tight':False}
            combined_plot(mass_plot,dl,"LO",p,**kargs,**kwargs)
            combined_plot(mass_plot,dl,"NLO",p,**kargs,**kwargs)
            combined_plot(mass_plot,dl,"NLO_PLUS_NLL",p,**kargs,**kwargs)
    elif scale:
        for i in [0,1]:
            kargs = {'logy':[True,False][i],'mask':dl["LO_SCALE"]!=np.array(None), 'axes':axs[i],'K':[False,True][i],'tight':False}
            mass_plot(dl,  "LO_SCALE",p,           **kargs,**kwargs,label="LO")
            mass_plot(dl,  "NLO_SCALE",p,          **kargs,**kwargs,label="NLO")
            mass_plot(dl,  "NLO_PLUS_NLL_SCALE",p, **kargs,**kwargs,label="NLO+NLL")
    elif pdf:
        for i in [0,1]:
            kargs = {'logy':[True,False][i],'mask':dl["LO_PDF"]!=np.array(None), 'axes':axs[i],'K':[False,True][i],'tight':False}
            mass_plot(dl,  "LO_PDF",p,           **kargs,**kwargs,label="LO")
            mass_plot(dl,  "NLO_PDF",p,          **kargs,**kwargs,label="NLO")
            mass_plot(dl,  "NLO_PLUS_NLL_PDF",p, **kargs,**kwargs,label="NLO+NLL")
    elif plehn:
        axs[0].set_ylim([0.2*10**-2,10**3])
        axs[1].set_ylim([0.9,1.85])
        for i in [0,1]:
            kargs = {'yscale':1000, 'yaxis':"$\\sigma$ [fb]",'logy':[True,False][i],'axes':axs[i],'K':[False,True][i],'tight':False,'error':False}
            if i == 0:
                mass_plot(dl,  "LO",p,           **kargs,**kwargs,data_color='b',fmt='-',label="LO"if i==0 else "")
            mass_plot(dl,  "RNLO",p,          **kargs,**kwargs,data_color='k',K_plus_1=True,fmt='-.',label="Real"if i==0 else "")
            if i == 0:
                mass_plot(dl,  "NLO",p,          **kargs,**kwargs,data_color='r',fmt='-',label="NLO"if i==0 else "")
            mass_plot(dl,  "VNLO_PLUS_P_PLUS_K",p,          **kargs,**kwargs,K_plus_1=True,data_color='k',fmt=':',label="Virtual"if i==0 else "")
            if i == 1:
                mass_plot(dl,  "RNLO_PLUS_VNLO_PLUS_P_PLUS_K",p,K_plus_1=True,          **kargs,**kwargs,data_color='r',label="NLO"if i==0 else "")
            #mass_plot(dl,  "NLO_PLUS_NLL",p, **kargs,**kwargs,label="NLO+NLL")


#TODO unit and yscale for each case and mass_and_plot!
def mass_and_ratio_plot(dl,li,p,scale=False,pdf=False,combined=False,cont = False,figsize=(6,4),plot_data=True,fill=True,unit="pb",yscale=1.0,ylim=None,**kwargs):
    global fig, axs
    if not cont:
        fig, axs = plt.subplots(2, 1, figsize=figsize, sharex=True, gridspec_kw={'height_ratios': [2, 1]})
        # Remove horizontal space between axes
        fig.subplots_adjust(hspace=0)
        title(axs[0],li[0],**kwargs)
    kinv = {'xaxis':"$M$ [GeV]",'yaxis':"$d\\sigma/dM$ ["+unit+"/GeV]"}
    if ylim is not None:
        axs[0].set_ylim(ylim)
    if combined:
        for i in [0,1]:
            kargs = {'logy' : [p!="invariant_mass",False][i], 'axes':axs[i],'tight':False}
            if p == "invariant_mass":
                plot(dl,  "invariant_mass","LO",           **kinv,**kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="LO"if i==0 else "")
                plot(dl,  "invariant_mass","NLO",          **kinv,**kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO"if i==0 else "")
                plot(dl,  "invariant_mass","NLO_PLUS_NLL", **kinv,**kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO+NLL"if i==0 else "")
                if i == 1:
                    plot(dl,  "invariant_mass","NLO_PLUS_NLL_OVER_NLO",**kinv, interpolate=True,plot_data=False,fill=False,**kargs,**kwargs,data_color='0',label="(NLO+NLL)/NLO")
            else:
                combined_plot(mass_plot,dl,"LO",p,plot_data=plot_data,fill=fill,ratio=[False,True][i],**kargs,**kwargs)
                combined_plot(mass_plot,dl,"NLO",p,plot_data=plot_data,fill=fill,ratio=[False,True][i],**kargs,**kwargs)
                combined_plot(mass_plot,dl,"NLO_PLUS_NLL",p,plot_data=plot_data,fill=fill,ratio=[False,True][i],**kargs,**kwargs)
                if i == 1:
                    mass_plot(dl,  "NLO_PLUS_NLL_OVER_NLO",p,interpolate=True,plot_data=False,fill=False,mask=dl["LO_SCALE"]!=np.array(None), *kargs,**kwargs,data_color='0',label="(NLO+NLL)/NLO")

    elif scale:
        for i in [0,1]:
            kargs = {'logy':[p!="invariant_mass",False][i],'mask':dl["LO_SCALE"]!=np.array(None), 'axes':axs[i],'tight':False}
            if p == "invariant_mass":
                plot(dl,  "invariant_mass","LO_SCALE",         **kinv,  **kargs,**kwargs,yscale=yscale,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="LO" if i==0 else "")
                plot(dl,  "invariant_mass","NLO_SCALE",       **kinv,   **kargs,**kwargs,yscale=yscale,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO"if i==0 else "")
                plot(dl,  "invariant_mass","NLO_PLUS_NLL_SCALE", **kinv,**kargs,**kwargs,yscale=yscale,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO+NLL"if i==0 else "")
                if i == 1:
                    plot(dl,  "invariant_mass","NLO_PLUS_NLL_OVER_NLO",**kargs, interpolate=True,plot_data=False,fill=False,xaxis="$M$ [GeV]",yaxis="Ratio",**kwargs,data_color='0',label="(NLO+NLL)/NLO")
            else:
                mass_plot(dl,  "LO_SCALE",p,           **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="LO"if i==0 else "")
                mass_plot(dl,  "NLO_SCALE",p,          **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO"if i==0 else "")
                mass_plot(dl,  "NLO_PLUS_NLL_SCALE",p, **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO+NLL"if i==0 else "")
                if i == 1:
                    mass_plot(dl,  "NLO_PLUS_NLL_OVER_NLO",p, interpolate=True,plot_data=False,fill=False,yaxis="Ratio",**kargs,**kwargs,data_color='0',label="(NLO+NLL)/NLO")
    elif pdf:
        for i in [0,1]:
            kargs = {'logy':[p!="invariant_mass",False][i],'mask':dl["LO_PDF"]!=np.array(None), 'axes':axs[i],'tight':False}
            if p == "invariant_mass":
                plot(dl,  "invariant_mass","LO_PDF",          **kinv, **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="LO"if i==0 else "")
                plot(dl,  "invariant_mass","NLO_PDF",         **kinv, **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO"if i==0 else "")
                plot(dl,  "invariant_mass","NLO_PLUS_NLL_PDF", **kinv,**kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO+NLL"if i==0 else "")
                if i == 1:
                    plot(dl,  "invariant_mass","NLO_PLUS_NLL_OVER_NLO",**kinv, interpolate=True,plot_data=False,fill=False,**kargs,**kwargs,data_color='0',label="(NLO+NLL)/NLO")
            else:
                mass_plot(dl,  "LO_PDF",p,           **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="LO"if i==0 else "")
                mass_plot(dl,  "NLO_PDF",p,          **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO"if i==0 else "")
                mass_plot(dl,  "NLO_PLUS_NLL_PDF",p, **kargs,**kwargs,plot_data=plot_data,fill=fill,ratio=[False,True][i],label="NLO+NLL"if i==0 else "")
                if i == 1:
                    mass_plot(dl,  "NLO_PLUS_NLL_OVER_NLO",p, interpolate=True,plot_data=False,fill=False,**kargs,**kwargs,data_color='0',label="(NLO+NLL)/NLO")



from distutils.log import warn
from enum import IntEnum
#from math import dist
import copy
import warnings
import numpy as np
from typing import List

import pyslha
from .util import get_LR_partner, lhapdf_name_to_id, namehash

import lhapdf

in_dir = "./input/"
"""input directory"""
out_dir = "./output/"
"""output directory"""
pre = "nice -n 5"
"""run command prefix."""



def get_input_dir():
    """
    Gets the input directory.
    """
    global in_dir
    return in_dir


def get_output_dir():
    """
    Gets the input directory.
    """
    global out_dir
    return out_dir


def set_input_dir(ind):
    """
    Sets the input directory.
    """
    global in_dir
    in_dir = ind


def set_output_dir(outd):
    """
    Sets the output directory.
    """
    global out_dir
    out_dir = outd


def set_pre(ppre):
    """
    Sets the command prefix. 
    """
    global pre
    pre = ppre


def get_pre():
    """
    Gets the command prefix. 
    """
    global pre
    return pre


class Order(IntEnum):
    """
    Computation orders.
    """
    LO = 0
    NLO = 1
    NLO_PLUS_NLL = 2


class Input:
    """
    Input for computation.
    """
    # TODO allow unspecified input? Maybe with kwargs + defaults
    def __init__(self, order :Order, energy, particle1: int, particle2: int, slha: str, pdf_lo: str, pdf_nlo: str, mu_f=1.0, mu_r=1.0, pdfset_lo=0, pdfset_nlo=0,precision=0.01,max_iters=50, invariant_mass="auto",result="total",pt="auto",id="",model_path="/opt/MG5_aMC_v2_7_0/models/MSSMatNLO_UFO"):
        self.order = order
        self.energy = energy
        self.energyhalf = energy/2.
        self.particle1 = particle1
        self.particle2 = particle2
        self.slha = slha
        self.pdf_lo = pdf_lo
        self.pdfset_lo = pdfset_lo
        self.pdf_nlo = pdf_nlo
        self.pdfset_nlo = pdfset_nlo
        self.pdf_lo_id = lhapdf_name_to_id(pdf_lo)
        self.pdf_nlo_id = lhapdf_name_to_id(pdf_nlo)
        self.mu_f = mu_f
        self.mu_r = mu_r
        self.precision = precision
        self.max_iters = max_iters
        self.invariant_mass = invariant_mass
        self.pt = pt
        self.result = result
        self.id = id
        self.model_path = model_path
        update_slha(self)

def update_slha( i:Input ):
    """
    Updates dependent parameters in Input `i`.

    Mainly concerns the `mu` value used by `madgraph`.
    """
    b = pyslha.read(get_input_dir() + i.slha,ignorenomass=True)
    try:
        i.mu = (abs(b.blocks["MASS"][abs(i.particle1)]) +
                   abs(b.blocks["MASS"][abs(i.particle2)]))/2.
    except:
        pass
        





def scan(l: List[Input], var: str, range) -> List[Input]:
    """
    Scans a variable `var` over `range` in `l`.
    """
    ret = []
    for s in l:
        for r in range:
            tmp = copy.copy(s)
            setattr(tmp, var, r)
            ret.append(tmp)
    return ret


def scale_scan(l: List[Input], range=3, distance=2.):
    """
    Scans scale by varying `mu_f` and `mu_r`.
    They take `range` values from 1/`distance` to `distance` in lograthmic spacing.
    Only points with `mu_f`=`mu_r` or `mu_r/f`=1 or `mu_r/f`=`distance` or `mu_r/f`=1/`distance` are returned.
    """
    ret = []
    for s in l:
        # not on error pdfs
        if s.pdfset_nlo == 0:
            tmp = scan([s], "mu_f", np.logspace(np.log10(1. /
                        distance), np.log10(distance), range))
            tmp = scan(tmp, "mu_r", np.logspace(np.log10(1. /
                        distance), np.log10(distance), range))
            for t in tmp:
                if t.mu_f == 1.0 or t.mu_r == 1.0 or t.mu_f == t.mu_r or t.mu_f == distance or t.mu_f == 1./distance or t.mu_r == distance or t.mu_r == 1./distance:
                    ret.append(t)

        else:
            ret.append(s)

    return ret

def seven_point_scan(l: List[Input]):
    """
    Scans scale by varying `mu_f` and `mu_r` by factors of two excluding relative factors of 4.
    """
    range=3
    distance=2.
    ret = []
    for s in l:
        # not on error pdfs
        if s.pdfset_nlo == 0 and s.mu_f == 1.0 and s.mu_r == 1.0:
            tmp = scan([s], "mu_f", np.logspace(np.log10(1. /
                        distance), np.log10(distance), range))
            tmp = scan(tmp, "mu_r", np.logspace(np.log10(1. /
                        distance), np.log10(distance), range))
            for t in tmp:
                if not ((t.mu_f == distance and  t.mu_r == 1./distance) or (t.mu_r == distance and  t.mu_f == 1./distance)):
                    ret.append(t)

        else:
            ret.append(s)

    return ret


def pdf_scan(l: List[Input]):
    """
    Scans NLO PDF sets. 

    The PDF sets are infered from `lhapdf.getPDFSet` with the argument of `pdfset_nlo`.
    """
    ret = []
    for s in l:
        # only central scale
        if s.mu_f == 1.0 and s.mu_r == 1.0:
            set = lhapdf.getPDFSet(s.pdf_nlo)
            for r in range(set.size):
                tmp = copy.copy(s)
                setattr(tmp, "pdfset_nlo", r)
                ret.append(tmp)
        else:
            ret.append(s)
    return ret

def change_where(l:List[Input], dicts : dict, **kwargs):
    """
    Applies the values of `dicts` if the key value pairs in `kwargs` agree with a member of the list `l`.

    The changes only applied to the matching list members.
    """
    ret = []
    for s in l:
        ok = True
        for k,v in kwargs.items():
            if getattr(s,k) != v:
                ok = False
        if ok:
            tmp = copy.copy(s)
            for k,v in dicts.items():
                setattr(tmp, k, v)
            ret.append(tmp)
        else:
            ret.append(s)
            
    return ret

def scan_invariant_mass(l : List[Input],diff,points,low=0.001):
    """
    Logarithmic `invariant_mass` scan close to the production threshold.
    """
    ret = []
    for s in l:
        for r in s.mu*2.+ low+ (np.logspace(np.log10(low),np.log10(1+low),points)-low) *diff:
            tmp = copy.copy(s)
            setattr(tmp, "invariant_mass", r)
            tmp.result = "m"
            ret.append(tmp)
    return ret

def mass_scan(l: List[Input], var: int, range, diff_L_R=None) -> List[Input]:
    """
    Scans the PDG identified mass `var` over `range` in the list `l`.
    `diff_L_R` allows to set a fixed difference between masses of left- and right-handed particles.
    """
    ret = []
    for s in l:
        for r in range:
            d = None
            try:
                d = pyslha.read(s.slha)
            except:
                d = pyslha.read(get_input_dir() + s.slha)
            d.blocks["MASS"][abs(var)] = r
            if not (diff_L_R is None):
                is_L, v = get_LR_partner(abs(var))
                d.blocks["MASS"][abs(v)] = r + is_L*diff_L_R

            newname = s.slha + "_mass_" + str(var) + "_" + str(r)
            pyslha.write(get_input_dir()+newname, d)

            tmp = copy.copy(s)
            setattr(tmp, "mass_" + str(var), r)
            setattr(tmp, "slha", newname)
            update_slha(tmp)
            ret.append(tmp)
    return ret

def slha_scan(l : List[Input],block,var,range : List) -> List[Input]:
    """
    Scan a generic 
    """
    return slha_scan_rel(l,lambda r,: [(block,var,r)],range)

def slha_scan_rel(l : List[Input],lambdas ,range : List) -> List[Input]:
    """
    Scan a generic slha variable.
    """
    ret = []
    for s in l:
        for r in range:
            d = None
            tmp = copy.copy(s)
            newname = s.slha
            try:
                d = pyslha.read(s.slha,ignorenomass=True)
            except:
                d = pyslha.read(get_input_dir() + s.slha,ignorenomass=True)
            ls = lambdas(r)
            for b,v,res in ls:
                d.blocks[b][v] = res
                setattr(tmp, b+ "_" + str(v), res)
                newname = newname +  "_" +str(b) + "_" + str(v) + "_" + str(res)
            pyslha.write(get_input_dir()+newname, d)

            setattr(tmp, "slha", newname)
            update_slha(tmp)
            ret.append(tmp)
    return ret


from typing import List
import subprocess
from string import Template
import numpy as np
import pkgutil

from hepi.util import namehash
from .. import Input, Result, LD2DL, get_output_dir, get_input_dir, get_pre
import re
from uncertainties import ufloat_fromstr
import os.path
from pathlib import Path
from .result import ResumminoResult, is_valid, parse_single
import enlighten
import time
import difflib
from smpl.parallel import par
import hashlib
import os 
import stat


resummino_path = "~/resummino/"
"""resummino folder containing the binary in './build/bin'."""



def set_path(p):
    """
    Set the path to the resummino folder containing the binary in './build/bin'.
    """
    global resummino_path
    resummino_path = p+ ("/" if p[-1]!="/" else "")


def get_path():
    """
    Returns the currently set resummino path
    """
    global resummino_path
    return resummino_path


class RunParams:
    """
    Parameters for Resummino.
    """
    def __init__(self, flags: str, in_path: str, out_path: str, skip=False):
        self.skip = skip
        self.flags = flags
        self.in_path = in_path
        self.out_path = out_path


def run(params: List[Input], noskip=False, bar=True, no_parse=False,para=True):
    """
    Run the passed list of parameters.
    """
    print("Running: " + str(len(params))  +" jobs" )
    rps = _queue(params, noskip)
    _run(rps, bar, no_parse,para)
    if not no_parse:
        outs = LD2DL(rps)["out_path"]
        results = _parse(outs)
        rdl = LD2DL(results)
        pdl = LD2DL(params)
        return {**rdl, **pdl}
    return {}


def _parse(outputs: List[str]) -> List[ResumminoResult]:
    """
    Parses Resummino output files and returns List of Results.
    """
    rsl = []
    for r in par(parse_single, outputs):
        rsl.append(r)
    return rsl



def _queue(params: List[Input], noskip=False) -> List[RunParams]:
    """
    Queues and generates Resummino run files.
    """
    global resummino_path
    Path("output").mkdir(parents=True, exist_ok=True)
    Path("input").mkdir(parents=True, exist_ok=True)
    ret = []
    for p in params:
        d = p.__dict__
        d["code"] = "RS"
        # TODO insert defautl if missing in d!
        name = namehash("_".join("".join(str(_[0]) + "_" + str(_[1]))
                        for _ in d.items()).replace("/", "-"))
        print(name)
        skip = False
        if not noskip and os.path.isfile(get_output_dir() + name + ".out") and is_valid(get_output_dir() + name + ".out",p,d):
            print("skip", end='')
            skip = True
        if not skip:
            data = pkgutil.get_data(__name__, "plot_template.in").decode(
                'utf-8')

            src = Template(data)
            result = src.substitute(d)
            open(get_input_dir() + name + ".in", "w").write(result)
            open(get_input_dir() + name + ".sh", "w").write("#!/bin/sh\n"+
                resummino_path + 'build/bin/resummino {} {} >> {}'.format(
                    get_input_dir()+name + ".in",["--lo", "--nlo", "--nll"][p.order],get_output_dir()+name + ".out"
                ))
            st = os.stat(get_input_dir() + name + ".sh")
            os.chmod(get_input_dir() + name + ".sh", st.st_mode | stat.S_IEXEC)
            open(get_output_dir() + name + ".out", "w").write(result + "\n\n")

            sname = d['slha']
            with open(get_input_dir() + sname, 'r') as f:
                #src = Template(f.read())
                #result = src.substitute(d)
                #open(get_input_dir() + sname + ".in", "w").write(result)
                open(get_output_dir() + name + ".out",
                     "a").write(f.read() + "\n\n")

        ret.append(RunParams(["--lo", "--nlo", "--nll"]
                             [p.order], get_input_dir()+name + ".in", get_output_dir()+name + ".out", skip))

    return ret


def _run(rps: List[RunParams], bar=True, no_parse=False,para=True):
    """
    Runs resummino per RunParam.
    """
    # TODO clean up on exit emergency
    global resummino_path
    # TODO RS build path checks?!?!
    template = get_pre() + " " +  "{}"

    # Run commands in parallel
    processes = []
    processesrpo = {}
    processesbar = {}
    manager = enlighten.get_manager()
    status_format = '{program}{fill} Stage: {stage}{fill} Status {status}:{fill}{lastline}'
    main_format = '{program}'
    mp = ""
    if bar:
        mp = rps[0].out_path
        main_bar = manager.status_bar(status_format=main_format,
                                      program=mp)
        mp = rps[1].out_path

    for rp in rps:
        if not rp.skip:
            command = template.format(rp.in_path.replace(".in",".sh"))
            process = subprocess.Popen(command, shell=True)
            processes.append(process)
            processesrpo[process] = rp.out_path
            if not para:
                process.wait()
            if no_parse:
                time.sleep(5)
            if bar:
                nnn = ""
                nn = ""
                ch = False
                for i, s in enumerate(difflib.ndiff(mp+"_", rp.out_path+"_")):
                    if s[-1] == "_":
                        if ch:
                            nnn = nnn + " ... " + nn
                        ch = False
                        nn = ""
                    if s[0] == '+':
                        ch = True
                    if (s[0] == '+' or s[0] == ' ') and s[-1] != "_":
                        nn = nn + s[-1]
                processesbar[process] = manager.status_bar(status_format=status_format,
                                                           program=nnn,
                                                           stage='INIT',
                                                           status='OKAY',
                                                           lastline="0")
        mp = rps[0].out_path
    if bar:
        c = True
        while c:
            if len(processes) > 0:
                time.sleep(10)
            c = False
            for p in processes:
                if p.poll() is None:
                    c = True
                    pat = re.compile(r'^\* (.*) \*')
                    n = ""
                    cl = 0
                    with open(processesrpo[p], mode="r") as f:
                        for l in f:
                            tmp = pat.search(l)
                            if(tmp is not None):
                                n = tmp.group(1)
                                cl = 0
                            cl = cl+1

                    processesbar[p].update(
                        stage=n, status=cl, lastline=l[int(len(l)/2)::])
                else:
                    processesbar[p].update(
                        stage="DONE", status='DONE', lastline="")
        for p in processes:
            processesbar[p].close()
        main_bar.close()

    if not no_parse:
        # Collect statuses
        output = [p.wait() for p in processes]
        return output
    return []



from hepi.input import Order
from .. import Input, Result, LD2DL, get_output_dir, get_input_dir
import re
from uncertainties import ufloat_fromstr
from string import Template
import pkgutil
import warnings
import os

class ResumminoResult(Result):
    """
    Resummino Result Data.
    """
    def __init__(self, lo, nlo, nlo_plus_nll, vnlo, p_plus_k, rnlog, rnloq):
        Result.__init__(self, lo, nlo, nlo_plus_nll)
        self.VNLO = vnlo
        self.P_PLUS_K = p_plus_k
        self.RNLOG = rnlog
        self.RNLOQ = rnloq
        if not (vnlo is None or p_plus_k is None):
            self.VNLO_PLUS_P_PLUS_K = self.VNLO + self.P_PLUS_K
        else:
            self.VNLO_PLUS_P_PLUS_K = None
        if not (rnlog is None or rnloq is None):
            self.RNLO = rnlog+rnloq
        else:
            self.RNLO = None
        if not ( self.RNLO is None or self.VNLO_PLUS_P_PLUS_K is None):
            self.RNLO_PLUS_VNLO_PLUS_P_PLUS_K = self.RNLO + self.VNLO_PLUS_P_PLUS_K
        else:
            self.RNLO_PLUS_VNLO_PLUS_P_PLUS_K = None

def is_valid(file:str,p:Input,d):
    """
    Verifies that an file is a complete output.
    """
    order = p.order
    data = pkgutil.get_data(__name__, "plot_template.in").decode(
                'utf-8')
    src = Template(data)
    result = src.substitute(d)
    sname = d['slha']
    with open(file,mode='r') as f:
        with open(get_input_dir() + sname, 'r') as sf:
            if not f.read().startswith(result + "\n\n" + sf.read()):
                #warnings.warn("Possible hash collision in " + file + " -> deleted",RuntimeWarning)
                os.remove(file)
                #exit(1)
                #return False
                print("HASH COLLIDED", file)
                return False
    res = parse_single(file)
    if res.LO is not None and order is Order.LO:
        return True
    if res.LO is not None and res.NLO is not None and order is Order.NLO:
        return True
    if res.LO is not None and res.NLO is not None and res.NLO_PLUS_NLL is not None and order is Order.NLO_PLUS_NLL:
        return True
    print("RESTART" ,res.LO, res.NLO,res.NLO_PLUS_NLL, file)
    return False


def parse_single(file) -> ResumminoResult:
    """
    Extracts LO, NLO and NLO+NLL from resummino output file.
    """
    # TODO generalize units like RS
    lo_pattern = re.compile(r'^LO = \((.*)\) pb')
    nlo_pattern = re.compile(r'^NLO = \((.*)\) pb')
    nll_pattern = re.compile(r'^NLO\+NLL = \((.*)\) pb')

    vnlo_pattern = re.compile(r'^vNLO = \((.*)\) pb')
    ppk_pattern = re.compile(r'^P\+K = \((.*)\) pb')
    rnlog_pattern = re.compile(r'^rNLOg = \((.*)\) pb')
    rnloq_pattern = re.compile(r'^rNLOq = \((.*)\) pb')

    lo_result = None
    nlo_result = None
    nll_result = None

    vnlo_result = None
    ppk_result = None
    rnlog_result = None
    rnloq_result = None
    with open(file) as output:
        for line in output:
            tmp = lo_pattern.search(line)
            if tmp is not None:
                lo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = nlo_pattern.search(line)
            if tmp is not None:
                nlo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = nll_pattern.search(line)
            if tmp is not None:
                nll_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = vnlo_pattern.search(line)
            if tmp is not None:
                vnlo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = ppk_pattern.search(line)
            if tmp is not None:
                ppk_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = rnlog_pattern.search(line)
            if tmp is not None:
                rnlog_result = ufloat_fromstr(
                    tmp.group(1).replace("+-", "+/-"))
            tmp = rnloq_pattern.search(line)
            if tmp is not None:
                rnloq_result = ufloat_fromstr(
                    tmp.group(1).replace("+-", "+/-"))
    return ResumminoResult(lo_result, nlo_result, nll_result, vnlo_result, ppk_result, rnlog_result, rnloq_result)
